#!/bin/sh
#
# check-dives.sh nr [nr ..]
# 
# AUTHOR: Dirk Hohndel <dirkhh at gmail dot com>
#         Copyright 2011
#         Licensed under the MIT License
#
# allows the automated extraction of dive and divelog information from the UEMIS Zurich SDA
# parses that information into XML files, and downloads matching XML files from the UEMIS 
# web site
# simply call with one or more dive numbers as arguments - these are dive numbers as seen 
# by the SDA - not the Dive IDs that are stored on the UEMIS web site
#
# the XML files will NOT be identical known issues at this stage
# 
# given various white space changes the best way to compare these files is
#	diff -w -u diveNNN.xml diveNNN.uemis.xml
#	diff -w -u divelogNNN.xml divelogNNN.uemis.xml
#
# diveNNN.xml:
#	- format of floats is not always consistent
#	- divesite information is missing (not extracted from SDA)
#	- gear information is missing (not extracted from SDA)
#
# divelogNNN.xml:
#	- date_time item isn't parsed
#	- various pressures are mis-calculated
#	- mod and tcd in the gas section aren't parsed
#
# even with these errors, divinglog should import the locally create XML files.
# if you find other inconsistencies (or want to help us by providing more sample
# data), please post all six corresponding files to the ScubaBoard thread:
# diveNNN.SDA divelogNNN.SDA diveNNN.xml divelogNNN.xml diveNNN.uemis.xml divelogNNN.uemis.xml
#

# set -x

user=""
pass=""
uemispath=""

if [ "$uemispath" == "" ]; then
	echo "You need to set the path to your mounted uemis filesystem"
	echo "(often /Volumes/UEMISSDA)"
	echo "and you username and password for the uemis website."
	echo "To do this please edit this shell script"
	exit 1;
fi

# first log in and get the cookies
echo "Log in to uemis website"
curl -c cookies.txt -L --data "Login=$user&Password=$pass&LoginButton=Login" http://www.uemis.com/en/user/login > /dev/null 2>&1
if [ $? -ne 0 ]; then
	echo "failed to log in"
	exit 1;
fi

for nr in "$@"
do
	echo "process dive $nr"
	if [ ! -f dive$nr.SDA ]; then
		echo "don't have dive$nr.SDA file; extracting from SDA"
		java -jar myuemis.jar -o dive$nr.SDA -f $uemispath -d $nr > /dev/null 2>&1
	fi
	if [ ! -f divelog$nr.SDA ]; then
		echo "don't have divelog$nr.SDA file; extracting from SDA"
		java -jar myuemis.jar -o divelog$nr.SDA -f $uemispath -s $nr > /dev/null 2>&1
	fi
	echo "creating local .xml files"
	./parse_uemis dive$nr.SDA
	remote=$(grep -1 remote dive$nr.SDA | tail -1 | sed 's/[^0-9]*//g')

	echo "download corresponding xml files (remote dive ID is $remote)"
	curl -b cookies.txt http://www.uemis.com/en/sda_logbook/data_src/$remote/divelog -o divelog$nr.uemis.xml
	curl -b cookies.txt http://www.uemis.com/en/sda_logbook/data_src/$remote/dive -o dive$nr.uemis.xml

done
