/*
 * parse_uemis.c
 *
 * UEMIS data "file_content" parser
 * AUTHOR:  Dirk Hohndel - Copyright 2011
 *
 * Licensed under the MIT license.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <regex.h>

/*
 * following code is based on code found in at base64.sourceforge.net/b64.c
 * AUTHOR:         Bob Trower 08/04/01
 * COPYRIGHT:      Copyright (c) Trantor Standard Systems Inc., 2001
 * NOTE:           This source code may be used as you wish, subject to
 *                 the MIT license. 
 */

#include <stdio.h>
#include <stdlib.h>

/* show the data fields that we don't understand, yet */
#define DEBUG_MISSING 0

/*
 * Translation Table to decode (created by Bob Trower)
 */
static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";

/*
 * decodeblock
 *
 * decode 4 '6-bit' characters into 3 8-bit binary bytes
 */
void decodeblock( unsigned char in[4], unsigned char out[3] )
{   
    out[ 0 ] = (unsigned char ) (in[0] << 2 | in[1] >> 4);
    out[ 1 ] = (unsigned char ) (in[1] << 4 | in[2] >> 2);
    out[ 2 ] = (unsigned char ) (((in[2] << 6) & 0xc0) | in[3]);
}

/*
 * decode
 *
 * decode a base64 encoded stream discarding padding, line breaks and noise
 */
void decode( uint8_t *inbuf, uint8_t *outbuf, int inbuf_len )
{
    uint8_t in[4], out[3], v;
    int i,len,indx_in=0,indx_out=0;
    
    while (indx_in < inbuf_len) {
        for (len = 0, i = 0; i < 4 && (indx_in < inbuf_len); i++ ) {
            v = 0;
            while ((indx_in < inbuf_len) && v == 0) {
                v = inbuf[indx_in++];
                v = ((v < 43 || v > 122) ? 0 : cd64[ v - 43 ]);
                if (v) 
                    v = ((v == '$') ? 0 : v - 61);
            }
            if (indx_in < inbuf_len) {
                len++;
                if (v) 
                    in[i] = (v - 1);
            }
            else 
                in[i] = 0;
        }
        if( len ) {
            decodeblock( in, out );
            for( i = 0; i < len - 1; i++ )
                outbuf[indx_out++] = out[i];
        }
    }
}

/*
 * parse the divelog binary data and convert into matching XML
 */
int
parse_divelog_binary(char *base64, FILE *outfd)
{
    int ret=0;
    char *start;
    int i,len,datalen;
    uint8_t *header;

    start = strstr(base64,"<bin format=\"base64\">") + 21;
    len = strstr(start,"</bin>") - start;
    datalen = (len/4 + 1)*3;
    if (datalen < 0x123+0x25) {
	fprintf(stderr,"suspiciously short data block\n");
	ret=1;
	goto bail;
    }
    header = malloc(datalen);
    decode(start,header,len);
    
    if (memcmp(header,"Dive\01\00\00",7)) {
	ret=1;
	printf("Missing Dive100 header\n");
	goto bail;
    }
    fprintf(outfd,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    fprintf(outfd,"<divelog xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" structure_id=\"1\">\n");
    fprintf(outfd,"\t<computer_id>%u</computer_id>\n",*(uint32_t *)(header+9));
    fprintf(outfd,"\t<dive_number>%u</dive_number>\n",*(uint16_t *)(header+7));
    fprintf(outfd,"\t<pre_dive>\n\t\t<system>\n\t\t\t<modules>0</modules>\n");
    fprintf(outfd,"\t\t\t<computer_type>%u</computer_type>\n",*(uint16_t *)(header+13));
    fprintf(outfd,"\t\t\t<computer_settings>%u</computer_settings>\n",*(uint16_t *)(header+15));
    fprintf(outfd,"\t\t\t<software_version>%u</software_version>\n",*(uint16_t *)(header+17));
    fprintf(outfd,"\t\t</system>\n");
    fprintf(outfd,"\t\t<water>%d</water>\n",*(uint8_t *)(header+19));
    fprintf(outfd,"\t\t<units>\n");
    fprintf(outfd,"\t\t\t<length>%d</length>\n",*(uint8_t *)(header+20));
    fprintf(outfd,"\t\t\t<volume>%d</volume>\n",*(uint8_t *)(header+21));
    fprintf(outfd,"\t\t\t<pressure>%d</pressure>\n",*(uint8_t *)(header+22));
    fprintf(outfd,"\t\t\t<temperature>%d</temperature>\n",*(uint8_t *)(header+23));
    fprintf(outfd,"\t\t\t<weight>%d</weight>\n",*(uint8_t *)(header+24));
    fprintf(outfd,"\t\t\t<time>%d</time>\n",*(uint8_t *)(header+25));
    fprintf(outfd,"\t\t\t<date>%d</date>\n",*(uint8_t *)(header+26));
    fprintf(outfd,"\t\t</units>\n");
    fprintf(outfd,"\t\t<profile>\n");
    fprintf(outfd,"\t\t\t<experience>%d</experience>\n",*(uint8_t *)(header+27));
    fprintf(outfd,"\t\t\t<gender>%d</gender>\n",*(uint8_t *)(header+28));
    fprintf(outfd,"\t\t\t<body_weight>%d</body_weight>\n",*(uint16_t *)(header+29));
    fprintf(outfd,"\t\t\t<body_height>%d</body_height>\n",*(uint16_t *)(header+31));
    fprintf(outfd,"\t\t</profile>\n");
    fprintf(outfd,"\t\t<dive_time>\n");
    fprintf(outfd,"\t\t\t<date_time></date_time>\n");
    fprintf(outfd,"\t\t\t<time_zone>%d</time_zone>\n",*(uint8_t *)(header+33));
    fprintf(outfd,"\t\t\t<summer_time>%d</summer_time>\n",*(uint8_t *)(header+34));
#if DEBUG_MISSING
    printf("unknown date_time %16lx\n",*(uint64_t *)(header+35));
#endif
    fprintf(outfd,"\t\t</dive_time>\n");
    fprintf(outfd,"\t\t<ambient>\n");
    fprintf(outfd,"\t\t\t<air_pressure>%d</air_pressure>\n",*(uint16_t *)(header+43));
    fprintf(outfd,"\t\t\t<temperature>%d</temperature>\n",*(uint16_t *)(header+45));
    fprintf(outfd,"\t\t\t<gps>\n");
    fprintf(outfd,"\t\t\t\t<longitude>0.000000</longitude>\n");
    fprintf(outfd,"\t\t\t\t<latitude>0.000000</latitude>\n");
    fprintf(outfd,"\t\t\t\t<altitude>0</altitude>\n");
    fprintf(outfd,"\t\t\t</gps>\n");
    fprintf(outfd,"\t\t</ambient>\n");
#if DEBUG_MISSING
    printf("unknown %8x\n",*(uint32_t *)(header+47));
    printf("unknown %8x\n",*(uint32_t *)(header+51));
    printf("unknown %4x\n",*(uint16_t *)(header+55));
#endif
    fprintf(outfd,"\t\t<physio>\n");
    fprintf(outfd,"\t\t\t<no_fly_time>%d</no_fly_time>\n",*(uint16_t *)(header+57));
    fprintf(outfd,"\t\t\t<no_dive_time>%d</no_dive_time>\n",*(uint16_t *)(header+59));
    fprintf(outfd,"\t\t\t<desaturation_time>%d</desaturation_time>\n",*(uint16_t *)(header+61));
    fprintf(outfd,"\t\t\t<allowed_altitude>%d</allowed_altitude>\n",*(uint16_t *)(header+63));
    fprintf(outfd,"\t\t\t<cns>%d</cns>\n",*(uint8_t *)(header+65));
    fprintf(outfd,"\t\t\t<interval_time>%d</interval_time>\n",*(uint16_t *)(header+66));
    fprintf(outfd,"\t\t</physio>\n");
    fprintf(outfd,"\t\t<battery>\n");
    fprintf(outfd,"\t\t\t<main_level>%d</main_level>\n",*(uint16_t *)(header+68));
    fprintf(outfd,"\t\t\t<flow_dir>%d</flow_dir>\n",*(int16_t *)(header+70));
    fprintf(outfd,"\t\t\t<sender_1>%d</sender_1>\n",*(uint8_t *)(header+72));
    fprintf(outfd,"\t\t\t<sender_2>%d</sender_2>\n",*(uint8_t *)(header+73));
    fprintf(outfd,"\t\t\t<sender_3>%d</sender_3>\n",*(uint8_t *)(header+74));
    fprintf(outfd,"\t\t</battery>\n");
    fprintf(outfd,"\t\t<safety>\n");
    fprintf(outfd,"\t\t\t<level>%d</level>\n",*(uint8_t *)(header+75));
    fprintf(outfd,"\t\t\t<settings>\n");
    fprintf(outfd,"\t\t\t\t<deco_level>%d</deco_level>\n",*(uint8_t *)(header+76));
    fprintf(outfd,"\t\t\t\t<deep_stop_level>%d</deep_stop_level>\n",*(uint8_t *)(header+80));
    fprintf(outfd,"\t\t\t\t<stop_time>%d</stop_time>\n",*(uint8_t *)(header+84));
    fprintf(outfd,"\t\t\t\t<stop_depth_pressure>%3.6f</stop_depth_pressure>\n",(*(uint16_t *)(header+89))/128.0);
    fprintf(outfd,"\t\t\t</settings>\n");
    fprintf(outfd,"\t\t\t<settings>\n");
    fprintf(outfd,"\t\t\t\t<deco_level>%d</deco_level>\n",*(uint8_t *)(header+77));
    fprintf(outfd,"\t\t\t\t<deep_stop_level>%d</deep_stop_level>\n",*(uint8_t *)(header+81));
    fprintf(outfd,"\t\t\t\t<stop_time>%d</stop_time>\n",*(uint8_t *)(header+85));
    fprintf(outfd,"\t\t\t\t<stop_depth_pressure>%3.6f</stop_depth_pressure>\n",*(uint16_t *)(header+93)/128.0);
    fprintf(outfd,"\t\t\t</settings>\n");
    fprintf(outfd,"\t\t\t<settings>\n");
    fprintf(outfd,"\t\t\t\t<deco_level>%d</deco_level>\n",*(uint8_t *)(header+78));
    fprintf(outfd,"\t\t\t\t<deep_stop_level>%d</deep_stop_level>\n",*(uint8_t *)(header+82));
    fprintf(outfd,"\t\t\t\t<stop_time>%d</stop_time>\n",*(uint8_t *)(header+86));
    fprintf(outfd,"\t\t\t\t<stop_depth_pressure>%3.6f</stop_depth_pressure>\n",*(uint16_t *)(header+97)/128.0);
    fprintf(outfd,"\t\t\t</settings>\n");
    fprintf(outfd,"\t\t\t<settings>\n");
    fprintf(outfd,"\t\t\t\t<deco_level>%d</deco_level>\n",*(uint8_t *)(header+79));
    fprintf(outfd,"\t\t\t\t<deep_stop_level>%d</deep_stop_level>\n",*(uint8_t *)(header+83));
    fprintf(outfd,"\t\t\t\t<stop_time>%d</stop_time>\n",*(uint8_t *)(header+87));
    fprintf(outfd,"\t\t\t\t<stop_depth_pressure>%3.6f</stop_depth_pressure>\n",*(uint16_t *)(header+101)/128.0);
    fprintf(outfd,"\t\t\t</settings>\n");
#if DEBUG_MISSING
    printf("unknown 88 %8x\n",*(uint32_t *)(header+88));
    printf("unknown 92 %8x\n",*(uint32_t *)(header+92));
    printf("unknown 96 %8x\n",*(uint32_t *)(header+96));
    printf("unknown 100 %8x\n",*(uint32_t *)(header+100));
    printf("unknown 104 %8x\n",*(uint32_t *)(header+104));
#endif
    fprintf(outfd,"\t\t</safety>\n");
    fprintf(outfd,"\t\t<limits>\n");
    /* missing the max_depth_pressure */
    fprintf(outfd,"\t\t\t<max_depth_pressure>3992.000000</max_depth_pressure>\n");
    fprintf(outfd,"\t\t\t<max_dive_time>%d</max_dive_time>\n",*(uint16_t *)(header+108));
    fprintf(outfd,"\t\t\t<max_deco_time>%d</max_deco_time>\n",*(uint16_t *)(header+110));
    fprintf(outfd,"\t\t\t<gas_warning_time_rgt>%d</gas_warning_time_rgt>\n",*(uint8_t *)(header+112));
    fprintf(outfd,"\t\t\t<divetime_info>%d</divetime_info>\n",*(uint8_t *)(header+113));
    fprintf(outfd,"\t\t\t</limits>\n");
    /* still gaps decpoding the gas section */
    /* size is incorrect for 4L tanks (shows 0) */
    fprintf(outfd,"\t\t<gas>\n");
    fprintf(outfd,"\t\t\t<template>1</template>\n");
    fprintf(outfd,"\t\t\t\t<air>\n");
    fprintf(outfd,"\t\t\t\t\t<bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+116)-0x40800000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+120));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+121)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+134))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t</air>\n");
    fprintf(outfd,"\t\t\t\t<nitrox_1>\n");
    fprintf(outfd,"\t\t\t\t\t<bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+141)-0x40800000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+145));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+146)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+159))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t</nitrox_1>\n");
    fprintf(outfd,"\t\t\t\t<nitrox_2>\n");
    fprintf(outfd,"\t\t\t\t\t<bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+166)-0x40800000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+170));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+171)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+184))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t<deco_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+191)-0x40400000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+195));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+196)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+209))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</deco_tank>\n");
    fprintf(outfd,"\t\t\t\t</nitrox_2>\n");
    fprintf(outfd,"\t\t\t\t<nitrox_3>\n");
    fprintf(outfd,"\t\t\t\t\t<bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+216)-0x40800000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+220));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+221)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+234))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t\t<deco_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+241)-0x40400000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+245));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+246)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+259))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</deco_tank>\n");
    fprintf(outfd,"\t\t\t\t\t<travel_tank>\n");
    fprintf(outfd,"\t\t\t\t\t\t<size>%2.6f</size>\n",(*(uint32_t *)(header+266)-0x40800000)/1048576.0);
    fprintf(outfd,"\t\t\t\t\t\t<oxygen>%d</oxygen>\n",*(uint8_t *)(header+270));
    fprintf(outfd,"\t\t\t\t\t\t<max_po2>%1.6f</max_po2>\n",(*(uint32_t *)(header+271)-0x3F000000)/8388608.0);
    fprintf(outfd,"\t\t\t\t\t\t<mod></mod>\n");
    fprintf(outfd,"\t\t\t\t\t\t<tcd></tcd>\n");
    fprintf(outfd,"\t\t\t\t\t\t<gas_limit>%4.6f</gas_limit>\n",(*(uint16_t *)(header+284))/32.0);
    fprintf(outfd,"\t\t\t\t\t\t<gas_reserve></gas_reserve>\n");
    fprintf(outfd,"\t\t\t\t\t</travel_tank>\n");
    fprintf(outfd,"\t\t\t\t</nitrox_3>\n");
    fprintf(outfd,"\t\t</gas>\n");
#if DEBUG_MISSING
    for (i=116;i<0x123;i+=25) {
	int ii;
	printf("unknown %d ",i);
	for (ii=0;ii<25;ii++)
	    printf(" %02x",*(uint8_t *)(header+i+ii));
	printf("\n");
    }
#else
    i=0x123;
#endif
    fprintf(outfd,"\t\t</pre_dive>\n");
    fprintf(outfd,"\t\t<dive>\n");
    while ((i < datalen) && (*(uint16_t *)(header+i))) {
	/* it seems that a dive_time of 0 indicates the end of the valid readings */
	fprintf(outfd,"\t\t\t<reading>\n");
	fprintf(outfd,"\t\t\t\t<dive_time>%d</dive_time>\n",*(uint16_t *)(header+i));
	fprintf(outfd,"\t\t\t\t<water_pressure>%d</water_pressure>\n",*(uint16_t *)(header+i+2));
	fprintf(outfd,"\t\t\t\t<dive_temperature>%d</dive_temperature>\n",*(uint16_t *)(header+i+4));
	fprintf(outfd,"\t\t\t\t<ascent_speed>%d</ascent_speed>\n",*(uint8_t *)(header+i+6));
	fprintf(outfd,"\t\t\t\t<work_fact>%d</work_fact>\n",*(uint8_t *)(header+i+7));
	fprintf(outfd,"\t\t\t\t<cold_fact>%d</cold_fact>\n",*(uint8_t *)(header+i+8));
	fprintf(outfd,"\t\t\t\t<bubble_fact>%d</bubble_fact>\n",*(uint8_t *)(header+i+9));
	fprintf(outfd,"\t\t\t\t<ascent_time>%d</ascent_time>\n",*(uint16_t *)(header+i+10));
#if DEBUG_MISSING
	printf("unknown 12 %04x %d\n",*(uint16_t *)(header+i+12),*(uint16_t *)(header+i+12));
#endif
	/* don't know where ascent_time_opt is encoded */
	fprintf(outfd,"\t\t\t\t<ascent_time_opt>0</ascent_time_opt>\n");
	fprintf(outfd,"\t\t\t\t<p_amb_tol>%d</p_amb_tol>\n",*(uint16_t *)(header+i+14));
	fprintf(outfd,"\t\t\t\t<satt>%d</satt>\n",*(uint16_t *)(header+i+16));
	fprintf(outfd,"\t\t\t\t<hold_depth>%d</hold_depth>\n",*(uint16_t *)(header+i+18));
	fprintf(outfd,"\t\t\t\t<hold_time>%d</hold_time>\n",*(uint16_t *)(header+i+20));
	fprintf(outfd,"\t\t\t\t<active_tank>%d</active_tank>\n",*(uint8_t *)(header+i+22));
	fprintf(outfd,"\t\t\t\t<tank_pressure>%d</tank_pressure>\n",*(uint16_t *)(header+i+23));
	fprintf(outfd,"\t\t\t\t<consumption>%d</consumption>\n",*(uint16_t *)(header+i+25));
	fprintf(outfd,"\t\t\t\t<rgt>%d</rgt>\n",*(uint8_t *)(header+i+27));
	fprintf(outfd,"\t\t\t\t<cns>%d</cns>\n",*(uint8_t *)(header+i+28));
	fprintf(outfd,"\t\t\t\t<warning_1>%u</warning_1>\n",*(uint32_t *)(header+i+29));
	fprintf(outfd,"\t\t\t\t<warning_2>%u</warning_2>\n",*(uint32_t *)(header+i+33));
	fprintf(outfd,"\t\t\t</reading>\n");
	i += 0x25;
    }
    fprintf(outfd,"\t\t</dive>\n");
    /* fill in the post_dive data */
#if DEBUG_MISSING
    {
	int k,j=i;
	while (i < datalen) {
	    printf("unknown %3d: ", i-j);
	    for(k=0;k<8 && i<datalen;i++,k++)
		printf("%02x ",*(uint8_t *)(header+i));
	    printf("\n");
	}
	i=j;
    }
#endif
    fprintf(outfd,"\t<post_dive>\n");
    fprintf(outfd,"\t\t<dive_time>\n");
    fprintf(outfd,"\t\t\t<date_time></date_time>\n");
    fprintf(outfd,"\t\t\t<time_zone>%d</time_zone>\n",*(uint8_t *)(header+i));
    fprintf(outfd,"\t\t\t<summer_time>%d</summer_time>\n",*(uint8_t *)(header+i+1));
#if DEBUG_MISSING
    printf("unknown date_time %16lx\n",*(uint64_t *)(header+i+2));
#endif
    fprintf(outfd,"\t\t</dive_time>\n");
    fprintf(outfd,"\t\t<ambient>\n");
    fprintf(outfd,"\t\t\t<air_pressure>%d</air_pressure>\n",*(uint16_t *)(header+i+10));
    fprintf(outfd,"\t\t\t<temperature>%d</temperature>\n",*(uint16_t *)(header+i+12));
    fprintf(outfd,"\t\t\t<gps>\n");
    fprintf(outfd,"\t\t\t\t<longitude>0.000000</longitude>\n");
    fprintf(outfd,"\t\t\t\t<latitude>0.000000</latitude>\n");
    fprintf(outfd,"\t\t\t\t<altitude>0</altitude>\n");
    fprintf(outfd,"\t\t\t</gps>\n");
    fprintf(outfd,"\t\t</ambient>\n");
    fprintf(outfd,"\t\t<physio>\n");
    fprintf(outfd,"\t\t\t<no_fly_time>%d</no_fly_time>\n",*(uint16_t *)(header+i+24));
    fprintf(outfd,"\t\t\t<no_dive_time>%d</no_dive_time>\n",*(uint16_t *)(header+i+26));
    fprintf(outfd,"\t\t\t<desaturation_time>%d</desaturation_time>\n",*(uint16_t *)(header+i+28));
    fprintf(outfd,"\t\t\t<allowed_altitude>%d</allowed_altitude>\n",*(uint16_t *)(header+i+30));
    fprintf(outfd,"\t\t\t<cns>%d</cns>\n",*(uint8_t *)(header+i+32));
    fprintf(outfd,"\t\t\t<interval_time>%d</interval_time>\n",*(uint16_t *)(header+i+33));
    fprintf(outfd,"\t\t</physio>\n");
    fprintf(outfd,"\t\t<battery>\n");
    fprintf(outfd,"\t\t\t<main_level>%d</main_level>\n",*(uint16_t *)(header+i+35));
    fprintf(outfd,"\t\t\t<flow_dir>%d</flow_dir>\n",*(int16_t *)(header+i+37));
    fprintf(outfd,"\t\t\t<sender_1>%d</sender_1>\n",*(uint8_t *)(header+i+39));
    fprintf(outfd,"\t\t\t<sender_2>%d</sender_2>\n",*(uint8_t *)(header+i+40));
    fprintf(outfd,"\t\t\t<sender_3>%d</sender_3>\n",*(uint8_t *)(header+i+41));
    fprintf(outfd,"\t\t</battery>\n");
    fprintf(outfd,"\t\t<dive_data>\n");
    fprintf(outfd,"\t\t\t<dive_duration>%d</dive_duration>\n",*(int16_t *)(header+i+42));
    fprintf(outfd,"\t\t\t<depth_max>%d</depth_max>\n",*(int16_t *)(header+i+44));
    fprintf(outfd,"\t\t\t<depth_avg>%d</depth_avg>\n",*(int16_t *)(header+i+46));
    fprintf(outfd,"\t\t\t<bottom_tank>\n");
    fprintf(outfd,"\t\t\t\t<pressure_start>%d</pressure_start>\n",*(int16_t *)(header+i+48));
    fprintf(outfd,"\t\t\t\t<pressure_end>%d</pressure_end>\n",*(int16_t *)(header+i+50));
    fprintf(outfd,"\t\t\t\t<cons_avg>%d</cons_avg>\n",*(int16_t *)(header+i+52));
    fprintf(outfd,"\t\t\t\t<cons_max>%d</cons_max>\n",*(int16_t *)(header+i+54));
    fprintf(outfd,"\t\t\t</bottom_tank>\n");
    fprintf(outfd,"\t\t\t<deco_tank>\n");
    fprintf(outfd,"\t\t\t\t<pressure_start>%d</pressure_start>\n",*(int16_t *)(header+i+56));
    fprintf(outfd,"\t\t\t\t<pressure_end>%d</pressure_end>\n",*(int16_t *)(header+i+58));
    fprintf(outfd,"\t\t\t\t<cons_avg>%d</cons_avg>\n",*(int16_t *)(header+i+60));
    fprintf(outfd,"\t\t\t\t<cons_max>%d</cons_max>\n",*(int16_t *)(header+i+62));
    fprintf(outfd,"\t\t\t</deco_tank>\n");
    fprintf(outfd,"\t\t\t<travel_tank>\n");
    fprintf(outfd,"\t\t\t\t<pressure_start>%d</pressure_start>\n",*(int16_t *)(header+i+64));
    fprintf(outfd,"\t\t\t\t<pressure_end>%d</pressure_end>\n",*(int16_t *)(header+i+66));
    fprintf(outfd,"\t\t\t\t<cons_avg>%d</cons_avg>\n",*(int16_t *)(header+i+68));
    fprintf(outfd,"\t\t\t\t<cons_max>%d</cons_max>\n",*(int16_t *)(header+i+70));
    fprintf(outfd,"\t\t\t</travel_tank>\n");
    fprintf(outfd,"\t\t\t<warning_2>%d</warning_2>\n",*(int32_t *)(header+i+72));
    fprintf(outfd,"\t\t\t<water_temp_max>%d</water_temp_max>\n",*(int16_t *)(header+i+76));
    fprintf(outfd,"\t\t\t<water_temp_min>%d</water_temp_min>\n",*(int16_t *)(header+i+78));
    fprintf(outfd,"\t\t</dive_data>\n");
    fprintf(outfd,"\t\t<deco/>\n");
    fprintf(outfd,"\t\t</post_dive>\n");

    /* now off to read the divelog */

#if DEBUG_MISSING
    for (i=114;i<0x123;i++)
	printf("unknown %02x\n",*(uint8_t *)(header+i));
#else
    i=0x123;
#endif
    
    fprintf(outfd,"</divelog>\n");
 bail:
    return ret;
}

int
simpleregex(char *buffer, char *regex, char *found) 
{
    int status;
    regex_t re;
    regmatch_t match[5];

    if (regcomp(&re, regex, 0) !=0) {
	fprintf(stderr,"internal error, regex failed!\n");
	exit(1);
    }
    status = regexec(&re,buffer,5,match,0);
    if(status == 0) {
	strncpy(found,buffer+match[1].rm_so,match[1].rm_eo-match[1].rm_so);
	found[match[1].rm_eo-match[1].rm_so] = '\0';
    }
    return(status == 0);
}

char * 
mygetline(FILE * f)
{
    size_t size = 0;
    size_t len  = 0;
    char * buf  = NULL;

    do {
        size += 1024;
        buf = realloc(buf,size);
        fgets(buf+len,1024,f);
        len = strlen(buf);
    } while (!feof(f) && buf[len-1]!='\n');
    return buf;
}

void
matchit(FILE *infd, FILE *outfd, char *regex, char *typeregex, char *outputformat)
{
    char *buffer;
    char found[1024]; /* DANGER - strcpy into a fixed length buffer */

    while (!feof(infd)) {
	buffer = mygetline(infd);
	if (simpleregex(buffer,regex,found)) {
	    buffer = mygetline(infd);
	    if (simpleregex(buffer,typeregex,found)) {
		fprintf(outfd,outputformat,found);
		break;
	    }
	}
    }
}

void
matchitarray(FILE *infd, FILE *outfd, char *regex, char *typeregex, char *outputformat, char **text)
{
    char *buffer;
    char found[1024]; /* DANGER - strcpy into a fixed length buffer */
    int indx;

    while (!feof(infd)) {
	buffer = mygetline(infd);
	if (simpleregex(buffer,regex,found)) {
	    buffer = mygetline(infd);
	    if (simpleregex(buffer,typeregex,found)) {
		/* here found is actually an index */
		sscanf(found,"%d",&indx);
		fprintf(outfd,outputformat,text[indx]);
		break;
	    }
	}
    }
}

/*
 * parse the two files extracted from the SDA
 * simply give the path to the diveNNNN.SDA file as argument
 * the program assumes a second file divelogNNNN.SDA in the same folder
 * and creates diveNNNN.xml and divelogNNNN.xml
 */
int
main(int argc, char **argv)
{
    char *base64=NULL;
    char *buffer;
    int indx=0,flag=0;
    struct tm date_time;
    FILE *divefile, *divelogfile, *divexmlfile, *divelogxmlfile;
    char *divelogfilename,*numptr;
    char *diveXMLname,*divelogXMLname;
    char found[1024]; /* DANGER - strcpy into fixed length buffer */

    if (argc != 2 || strstr(argv[1],"dive") == NULL || strstr(argv[1],".SDA") != argv[1]+strlen(argv[1])-4) {
	fprintf(stderr,"Usage: %s diveXXXX.SDA\n",argv[0]);
	fprintf(stderr,"diveXXXX.SDA and divelogXXXX.SDA are the two files extracted from SDA\n\n\n");
	exit(1);
    }
    divelogfilename = malloc(strlen(argv[1])+4);
    if (divelogfilename==NULL) {
	fprintf(stderr,"Out of memory\n");
	exit(1);
    }
    /* quick hack to create the divelog filename - there's better ways to do this */
    strcpy(divelogfilename,argv[1]);
    numptr = rindex(divelogfilename,'.');
    while (*numptr != 'e')
	numptr--;
    numptr++;
    memmove(numptr+3,numptr,strlen(numptr));
    strncpy(numptr,"log",3);
    diveXMLname = malloc(strlen(argv[1]));
    divelogXMLname = malloc(strlen(divelogfilename));
    if (diveXMLname == NULL || divelogXMLname == NULL) {
	fprintf(stderr,"Out of memory\n");
	exit(1);
    }
    strcpy(diveXMLname, argv[1]);
    strcpy(divelogXMLname,divelogfilename);
    strncpy(rindex(diveXMLname,'.'),".xml",4);
    strncpy(rindex(divelogXMLname,'.'),".xml",4);
    printf("%s %s\n",diveXMLname,divelogXMLname);

    divefile = fopen(argv[1],"r");
    divelogfile = fopen(divelogfilename,"r");
    if (divefile == NULL || divelogfile == NULL) {
	fprintf(stderr,"Can't open the input files %s and %s\n",
		argv[1],divelogfilename);
	exit(1);
    }
    divexmlfile = fopen(diveXMLname,"w");
    divelogxmlfile = fopen(divelogXMLname,"w");
    if (divefile == NULL || divelogfile == NULL) {
	fprintf(stderr,"Can't open the output files %s and %s\n",
		diveXMLname,divelogXMLname);
	exit(1);
    }
    fprintf(divexmlfile,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
    while (!feof(divefile)) {
	buffer = mygetline(divelogfile);
	if (simpleregex(buffer,"<hash.*dive.*version=\"\\([0-9.]*\\)\">",found)) {
	    fprintf(divexmlfile,"<dive version=\"%s\" locale=\"en\">\n",found);
	    break;
	}
    }
    matchit(divefile,divexmlfile,"<val key=\"dive_no\">", "<int>\\([0-9]*\\)</int>",
	    "\t<dive_no label=\"Dive Number\" nativetype=\"int\">%s</dive_no>\n");
    matchit(divefile,divexmlfile,"<val key=\"date\">", "<ts>\\(.*\\)</ts>",
	    "\t<date label=\"Date\" nativetype=\"ts\">\n\t\t<![CDATA[%s]]>\n\t</date>\n");
    matchit(divefile,divexmlfile,"<val key=\"duration\">", "<float>\\([0-9.]*\\)</float>",
	    "\t<duration label=\"Duration\" nativetype=\"float\">%s</duration>\n");
    matchit(divefile,divexmlfile,"<val key=\"depth\">", "<int>\\([0-9]*\\)</int>",
	    "\t<depth label=\"Depth\" nativetype=\"int\">%s</depth>\n");
    matchit(divefile,divexmlfile,"<val key=\"altitude\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<altitude label=\"Air Pressure (h Pa)\" nativetype=\"int\">%s</altitude>\n");
    matchit(divefile,divexmlfile,"<val key=\"decoindex\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<decoindex label=\"Decoindex\" nativetype=\"int\">%s</decoindex>\n");
    matchit(divefile,divexmlfile,"<val key=\"consumption\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<consumption label=\"Consumption\" nativetype=\"int\">%s</consumption>\n");
    matchitarray(divefile,divexmlfile,"<val key=\"gas\">", "<int>\\([0-9]*\\)</int>",
		 "\t<gas label=\"Gas\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</gas>\n",
		 (char*[]){"Air","Nitrox"});
    matchitarray(divefile,divexmlfile,"<val key=\"water\">", "<int>\\([0-9]*\\)</int>",
		 "\t<water label=\"Water Type\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</water>\n",
		 (char*[]){"Freshwater","Seawater"});
    matchitarray(divefile,divexmlfile,"<val key=\"divetype\">", "<int>\\([0-9]*\\)</int>",
		 "\t<divetype label=\"Dive Type\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</divetype>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchitarray(divefile,divexmlfile,"<val key=\"diveactivity\">", "<int>\\([0-9]*\\)</int>",
		 "\t<diveactivity label=\"Dive Activity\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</diveactivity>\n",
		 (char*[]){"-","-","-","-","-","-"});
    fprintf(divexmlfile,"\t<divespot/>\n");
    fprintf(divexmlfile,"\t<notes label=\"Notes\" nativetype=\"string\">\n\t\t<![CDATA[]]>\n\t</notes>\n");
    fprintf(divexmlfile,"\t<conditions label=\"Conditions\">\n\t\t<item>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Clouds\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Clouds label=\"Clouds\" nativetype=\"int\">%s</u8Clouds>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Deposit\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Deposit label=\"Precipitation\" nativetype=\"int\">%s</u8Deposit>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Wind\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Wind label=\"Wind\" nativetype=\"int\">%s</u8Wind>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Waves\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Waves label=\"Waves\" nativetype=\"int\">%s</u8Waves>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Current\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Current label=\"Current\" nativetype=\"int\">%s</u8Current>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8Visibility\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8Visibility label=\"Visibility\" nativetype=\"int\">%s</u8Visibility>\n");
    fprintf(divexmlfile,"\t\t</item>\n\t</conditions>\n");
    fprintf(divexmlfile,"\t<gear label=\"Gear\">\n\t\t<item>\n");
    matchitarray(divefile,divexmlfile,"<val key=\"u8DiveSuit\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8DiveSuit label=\"Dive Suit\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8DiveSuit>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchitarray(divefile,divexmlfile,"<val key=\"u8DiveSuitType\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8DiveSuitType label=\"Divesuit Type\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8DiveSuitType>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchitarray(divefile,divexmlfile,"<val key=\"u8SuitThickness\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8SuitThickness label=\"Suit Thickness\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8SuitThickness>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchit(divefile,divexmlfile,"<val key=\"f32Weight\">", "<float>\\([0-9.]*\\)</float>",
	    "\t<f32Weight label=\"f32Weight\" nativetype=\"float\">%s</f32Weight>\n");
    matchitarray(divefile,divexmlfile,"<val key=\"u8BCD\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8BCD label=\"BCD\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8BCD>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchitarray(divefile,divexmlfile,"<val key=\"u8Light\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8Light label=\"Light\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8Light>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchitarray(divefile,divexmlfile,"<val key=\"u8Camera\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8Camera label=\"Camera\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8Camera>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchit(divefile,divexmlfile,"<val key=\"u8BottomTankType\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8BottomTankType label=\"Bottom Tank Type\" nativetype=\"int\">%s</u8BottomTankType>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8TravelTankType\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8TravelTankType label=\"Travel Tank Type\" nativetype=\"int\">%s</u8TravelTankType>\n");
    matchit(divefile,divexmlfile,"<val key=\"u8DecoTankType\">", "<int>\\([0-9-]*\\)</int>",
	    "\t<u8DecoTankType label=\"Deco Tank Type\" nativetype=\"int\">%s</u8DecoTankType>\n");
    matchitarray(divefile,divexmlfile,"<val key=\"u8ScubaSet\">", "<int>\\([0-9]*\\)</int>",
		 "\t<u8ScubaSet label=\"Scuba Set\" nativetype=\"int\">\n\t\t<![CDATA[%s]]>\n\t</u8ScubaSet>\n",
		 (char*[]){"-","-","-","-","-","-"});
    matchit(divefile,divexmlfile,"<val key=\"u8Other\">", "<string>\\([^<]*\\)</string>",
	    "\t<u8Other label=\"Other\" nativetype=\"string\">\n\t\t<![CDATA[%s]]>\n\t</u8Other>\n");
    fprintf(divexmlfile,"\t\t</item>\n\t</gear>\n");
    fprintf(divexmlfile,"\t<buddies label=\"Buddies\">\n\t</buddies>\n\t</dive>\n");

    fclose(divexmlfile);

    while (!feof(divelogfile)) {
	buffer = mygetline(divelogfile);
	if (flag == 1 || strstr(buffer,"<bin format=\"base64\">")) {
	    base64 = realloc(base64,indx+strlen(buffer)+1);
	    strncpy(base64+indx,buffer,strlen(buffer)+1);
	    indx += strlen(buffer);
	    if(strstr(base64,"</bin>") == NULL)
		flag=1;
	    else {
		if (parse_divelog_binary(base64,divelogxmlfile)) {
		    fprintf(stderr,"failed to parse header\n");
		    exit(1);
		}
		indx=0;
		flag=0;
	    }
	} else if (strstr(buffer,"</bin>")) {
	    base64 = realloc(base64,indx+strlen(buffer)+1);
	    strncpy(base64+indx,buffer,strlen(buffer)+1);
	    indx=0;
	    flag=0;
	    if (parse_divelog_binary(base64,divelogxmlfile)) {
		fprintf(stderr,"failed to parse header\n");
		exit(1);
	    }
	}	    
	free(buffer);
    }

    fclose(divelogxmlfile);
    exit(0);
}
